/*!
 * javaRunner.js
 * author: Brent M. Dingle
 * last modified: 2015
 * Example html5 canvas game --- still in development (alpha version)
 *
 * All canvas elements are to be resized with the screen
 * -- for details see the 'window handling' stuff at the bottom of this file
 */

// ----------------------------------------------------------------------------
//                                                          GLOBAL VARs
// ----------------------------------------------------------------------------
// The below must match the id's used in the HTML file
// ----------------------------------------------------------------------------
var CANVAS_MENU_ID      = "canvasMenu";
var CANVAS_CREDITS_ID   = "canvasCredits";
var CANVAS_GAME_ID      = "canvasGame";
var PLAYAGAIN_BUT_ID    = "playAgainButton";
var GAMEOVER_DIV_ID     = "gameOverDiv";
var GAMEOVER_OVERLAY_ID = "gameOverOverlay";

var CANVAS_WIDTH      = 700;
var CANVAS_HEIGHT     = 240;

var OK_MIN_SCREEN_RATIO = 2.91;  // round down CANVAS_WIDTH / CANVAS_HEIGHT
var OK_MAX_SCREEN_RATIO = 2.93;  // round up CANVAS_WIDTH / CANVAS_HEIGHT


// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// theGame Object											theGame Object
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
var theGame = 
{
	// ----------------------------------------------------------------------------
	// Pseudo-constants
	// ----------------------------------------------------------------------------
	STATE_MENU:      0,
	STATE_CREDITS:   1,
	STATE_PLAYING:   5,
    STATE_GAMEOVER:  999,
    
    // Player Animation frames 
    // The second to last frame is jump up, and the last frame is jump down
    SPRITES: ['A5.1VEg254m6','A5..UoS201XU','41UfpSMDAg2+','MC05WlC45WM2'],

	
	// ----------------------------------------------------------------------------
	// Variables
	// ----------------------------------------------------------------------------
	gameState: 5,       // originally: 0 = menu, 1 = credits, 5 = playingGame
	isInitialized: 0,   // 0 = false, indicates if Init() function has been called
    
    startTime: 0,
    curTime:   0,
    prevTime:  0,
    delay:     0.03,

    // ------------------ more game specific items
    width:  CANVAS_WIDTH,
    height: CANVAS_HEIGHT,
    font:   '16px sans',
    colorStop1: '#83B5DB',
    colorStop2: '#DCAA84',
    gradient: null,
    
    lastScore: 0,
    topScore:  0,
    curScore:  0,   // score relates to number of frames/updates survived
    
    playerVel: 0,
    inAir:     1,   // =0 if we touch the ground
    jumpLeft:  0,   // how much 'time' is left for the current jump
    
    // playerX:  40,
    playerY: 150,    // about half the height
    speed: 10,      // game speed 
    
    sFrame: 0,
    
    bldgs: [],
    
	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------
	Init: function ()
	{
		// Init Local Canvas variables
		var gameCanvas = document.getElementById(CANVAS_GAME_ID);
		var gameCTX = gameCanvas.getContext("2d");
        gameCTX.font = this.font;
        
        this.height = gameCanvas.height;
        this.width  = gameCanvas.width;
        
        this.gradient = gameCTX.createLinearGradient(0, 0, 0, this.height);
        this.gradient.addColorStop(0, this.colorStop1);
        this.gradient.addColorStop(1, this.colorStop2);
        
        // html must define a button with id: playAgainButton
        // For mobile may also have to add a mouse touch event listener -- this 'click' only worked ok with iPad
        document.getElementById(PLAYAGAIN_BUT_ID).addEventListener('click', 
                                                function(evt) 
                                                {
                                                    theGame.ResetPlay();
                                                    evt.cancelBubble = true; // prevents click from going elsewhere after this
                                                } );
        
        theGame.ResetPlay();
        
        this.isInitialized = 1;
    },
    
	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------
    ProcessGameOver: function()
    {
        // quick and dirty game over
        //alert(this.curScore);
        //theGame.ResetPlay();
        
        // Note: fancier game over requires proper html and css setup
        theGame.GameOver();
    },
    
	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------
    GameOver: function()
    {
        // Display the game over div
        document.getElementById(GAMEOVER_DIV_ID).style.display = 'block';
        document.getElementById(GAMEOVER_OVERLAY_ID).style.display = 'block';
        
        // optionally disable the canvas
        //document.getElementById(CANVAS_GAME_ID).style.display = 'none';

        // Set game state
        this.gameState = this.STATE_GAMEOVER;
    },
    
	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------
    ResetPlay: function()
    {
        // Hide the game over div
        document.getElementById(GAMEOVER_DIV_ID).style.display = 'none';
        document.getElementById(GAMEOVER_OVERLAY_ID).style.display = 'none';

        // enable the canvas
        //document.getElementById(CANVAS_GAME_ID).style.display = 'block';
        
        // set state to playing
        this.gameState = this.STATE_PLAYING;
        
        dateObj  = new Date();
        this.startTime = dateObj.getTime();
        this.curTime = this.startTime;
        this.prevTime = this.startTime;
        
        this.lastScore = Math.round(this.curScore / 50);
        this.topScore  = Math.max(this.topScore, this.lastScore);
        this.curScore  = 0;
        this.playerVel = 0;
        this.bldgs     = [ {x:0, y:150, w:this.width}, {x:0, y:-200, w:this.width, n:1}];

        this.playerVel = 0;   // vertical velocity
        this.inAir     = 1;   // =0 if we touch the ground
        this.jumpLeft  = 0;   // how much 'time' is left for the current jump
        
        // this.playerX = 40; // keep player position static in x-dir
        this.playerY = 120; //150;   // about half the height
        this.speed   = 5; //10;    // game speed = player horizontal velocity
        
        this.sFrame  = 0;   // index into SPRITES
    },

	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------
    processKeyDown: function(keydown)
    {
        // Control the jump
        if (this.inAir == 0)  // if NOT in air, then can reset jump 'time' left
        {
            this.jumpLeft  = 5;
        }
    },
    
	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------
    processKeyUp: function(keyup)
    {
        this.jumpLeft  = 0;   // allow immediate fall
    },

	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------
    myRand: function(n,x)
    {
      return Math.round( Math.random() * (x-n)) + n;
    },
    
	// ----------------------------------------------------------------------------
    // This was originally designed to run at a rate of 30/1000 seconds per frame
    // (i.e. 1000/30 = 33 frames per second)
	// ----------------------------------------------------------------------------
    UpdatePlay: function()
    {
		var gameCanvas = document.getElementById(CANVAS_GAME_ID);
		var gameCTX = gameCanvas.getContext("2d");
        
        // clear the canvas
        gameCTX.fillStyle = this.gradient;
        gameCTX.fillRect(0, 0, this.width, this.height);
        
        // Set inAir to non-zero  (becomes 0 if player touches ground)
        this.inAir     = 1;
        
        // Update player position
        if (this.jumpLeft>0)
        {
            this.playerVel -= 3;
            this.inAir     = 1;  // may be redundant here
            this.jumpLeft--;
        }
        
        this.playerVel++;
        this.playerY += this.playerVel;
        
        // move and draw the buildings
        for(b in this.bldgs)
        {
            b = this.bldgs[b];
            b.x -= this.speed;   // game speed

            gameCTX.fillStyle = (b.b) ? '#744':'#000';
            gameCTX.fillRect(b.x, b.y, b.w, this.height);

            // Check if player is colliding   ----------------- player X hardcoded at 40
            if (40>b.x && 40<b.x+b.w && this.playerY>b.y && this.playerY<b.y+this.height) 
            {
                // Are we on the top/ground
                if (this.playerY - this.playerVel <= b.y)
                {
                    this.playerY = b.y;
                    this.inAir = 0;
                    this.playerVel = 0;
                }
                // or colliding?
                else
                {
                    if (!b.b)  // hit a building
                    {
                        //theGame.ResetPlay();  have better now
                        theGame.ProcessGameOver();
                        break;
                    }
                    else // hit an obstacle so slow down
                    {
                        this.speed *= 0.8;
                    }
                }
            }
            // If off bottom, stop
            if (this.playerY > this.height)
            {
                //theGame.ResetPlay();  have better now
                theGame.ProcessGameOver();
                break;
            }

            // Check if we need to create a new building
            //if (!b.n && b.x+b.w < 970) // ----------------------- hard coded width of world
            if (!b.n && b.x+b.w < this.width - 6)
            {
                this.bldgs.push(n = { x: b.n = this.width + theGame.myRand(0, 20 * this.speed),
                                      y: b.y - theGame.myRand(-15, 25),   // this needs fixed to not run heights off screen
                                      w: theGame.myRand(200, 999)} );
                                      
                // Sometimes we want a roof... (myRand(...)==0)
                if ( !theGame.myRand(0,10)) 
                {
                    this.bldgs.push( {x:n.x,
                                      y:n.y-420,
                                      w:n.w,
                                      n:1} );
                }
                // ...and sometimes up to 3 boxes. 
                // Add to the front of the array so the tallness is drawn behind the buildings
                for (z = theGame.myRand(-3,3); z-->0;) 
                {
                    this.bldgs.unshift( {x: theGame.myRand(n.x, n.x+n.w-20),
                                         y: n.y-20,
                                         w: 20,
                                         n: 1, b: 1} );
                }
            } // end if new building
            
        } // end for b in bldgs
        
        // Update speed and score stuff
        this.speed += 0.05;
        this.curScore += this.speed;
        
        // Draw player
        //gameCTX.fillRect(40, this.playerY-10, 10, 10);  <-- can use simple square
        this.sFrame++;
        this.sFrame %= 2; // only 2 frames for running
        if (this.inAir > 0)  // if not touching ground display jump pose
        {
            this.sFrame = 2;
        }
        var sprite = this.SPRITES[this.sFrame]; // animation frame
		i = 72;
		while (i) 
        {
			if (i-- % 6 === 0) 
            {
				// temp1 = base64 code for current sprite char
				temp1 = sprite.charCodeAt(i/6);
				// temp2 = unencoded base64 value
				temp2 = temp1 < 44 ? 0 : temp1 < 47 ? 1 : temp1 < 58 ? temp1 - 46 : temp1 < 91 ? temp1 - 53 : temp1 - 59;
			}
			if (temp2 % 2) 
            {
				gameCTX.fillRect(40 - 7 + (i % 7),
					            this.playerY-10 + Math.floor(i / 7), 1, 1);
			}
			temp2 >>>= 1;
		}
        
        // Draw the score info
        gameCTX.fillText('Hi ' + this.topScore 
                         + 'm Last ' + this.lastScore 
                         + 'm | ' + Math.round(this.curScore/50) + 'm', this.width-225, 20);
    },
    
	// ----------------------------------------------------------------------------
	// Update Based on this.gameState
	// ----------------------------------------------------------------------------
	Update: function ()
	{
		if (this.gameState == this.STATE_PLAYING)
		{
            // Only really want the game to update at about 33 FPS
            newDate  = new Date();
            this.curTime  = newDate.getTime();  // end time in milliseconds
            deltaTime = (this.curTime - this.prevTime) / 1000.0;    // delta time in seconds
            
            if (deltaTime >= this.delay)
            {
                this.prevTime = this.curTime;
                theGame.UpdatePlay();
            }
		}
    },
    
	// ----------------------------------------------------------------------------
	// Draw Based on this.gameState
	// Note: Draw should be called BEFORE update in runGame()
	// ----------------------------------------------------------------------------
	Draw: function ()
	{
		//console.log("DEBUG: Draw called, gameState = " + this.gameState );
		
		if (this.gameState == this.STATE_PLAYING)
		{
			//theGame.Draw();    <--------- TODO separate draw and update if possible
		}
		// else if (this.gameState == this.STATE_MENU)
		// {
			// theGame.DrawMenu();
		// }
		else
		{
			console.log("ERROR: Draw() -- unknown gameState = " + this.gameState );
		}
	},

	
};   // end of theGame variable/object

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// Window Handling Stuff
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
window.addEventListener("resize", doResize, false);
window.addEventListener("keydown", doKeyDown, false);
window.addEventListener("touchstart", doTouchStart, false);
window.addEventListener("keyup", doKeyUp, false);
window.addEventListener("touchend", doTouchEnd, false);

// ----------------------------------------------------------------------------
//                                                             doResize
//
// Need to know the canvas percentage of window for additional measure scaling
// if only have 80% of the height then need to scale to window.height*0.8
// i.e. other elements in the window are not scaled here
// OR must scale everything on the window
// ----------------------------------------------------------------------------
function doResize() {

	//var canvas1 = document.getElementById(CANVAS_MENU_ID);
	//var canvas2 = document.getElementById(CANVAS_CREDITS_ID);
	var canvas3 = document.getElementById(CANVAS_GAME_ID);
    
    var introText = document.getElementById("introText");
	
	// Size things
	//UpdateCanvas(canvas1);
	//UpdateCanvas(canvas2);
	UpdateCanvas(canvas3, introText);

	// If game has been initialized then force a game redraw
	// Note the game itself is NOT updated here... just changed the size of view
	if (theGame.isInitialized != 0)
	{
		//theGame.Draw();
        theGame.Update();
	}
};

// ----------------------------------------------------------------------------
//                                                               UpdateCanvas
// ----------------------------------------------------------------------------
function UpdateCanvas(canvas, introText) {
	
	//console.log("UpdateCanvas Called");
	
	// Here's the tricky part 
	// - Set the width and height of the canvas's coordinate system
	// - if not specified it defaults to width=300 and height=150
	// This is the size you "pretend" your world is (i.e. optimal)
	// 
	canvas.width = CANVAS_WIDTH;   //976;   // < ------------ ************************ WORLD SIZE HERE !!!
	canvas.height = CANVAS_HEIGHT; //336;   //   change these change the ratio check in the if below
	
	// DEBUG: console.log(canvas.width + " " + canvas.height);
	// DEBUG: console.log("W " + window.innerWidth + " " + window.innerHeight);
	// DEBUG: console.log("text height = " + introText.clientHeight);
    
	// Now we consider the actual display (the style width and height)
	// -- this is the size of the window you are using to view your world

    // canvas.style.width = window.innerWidth + 'px';       // This line and the next resize things
    // canvas.style.height = window.innerHeight + 'px';     // but do not maintain scale ratio
	
	// The below scale the game and maintain a ratio of 800x480 
	// by "adding" empty space to the right and bottom of the canvas as appropriate
	// You may need to put a box around the canvas to better see this
	
	var gameWidth = window.innerWidth;
	var gameHeight = window.innerHeight - introText.clientHeight;
	var scaleToFitX = gameWidth / canvas.width;   // canvas.width and height just set above
	var scaleToFitY = gameHeight / canvas.height;

	var currentScreenRatio = gameWidth / gameHeight;
	var optimalRatio = Math.min(scaleToFitX, scaleToFitY);

	canvas.style.position = "fixed";
	
    // For ref: width = 800, height = 450 then 800/450 ~= 1.78, bound by 1.77 and 1.79
	if (currentScreenRatio >= OK_MIN_SCREEN_RATIO && currentScreenRatio <= OK_MAX_SCREEN_RATIO) 
    {
		canvas.style.width = gameWidth + "px";
		canvas.style.height = gameHeight + "px";
		canvas.style.left = 0 + "px";
	}
	else 
    {
		canvas.style.width = canvas.width * optimalRatio + "px";
		canvas.style.height = canvas.height * optimalRatio + "px";
		canvas.style.left = ( (gameWidth - (canvas.width * optimalRatio)) / 2 ) + "px";

	}
};

// ----------------------------------------------------------------------------
//                                                                doKeyDown
//
// Setup things here so ESCAPE key will always return viewer to
// the menu screen.
// May also handle user keyboard input here (note canvas is not full screen)
//
// Press a key = jump
// ----------------------------------------------------------------------------
function doKeyDown(e)
{ 
    //console.log("DEBUG: doKeyDown entered")

	// if (e.keyCode == 27)
	// {
		// var menuCanvas = document.getElementById(CANVAS_MENU_ID);
		// var gameCanvas = document.getElementById(CANVAS_GAME_ID);
		// gameCanvas.style.display="none";
		// menuCanvas.style.display="block";
		// theGame.gameState = theGame.STATE_MENU;
	// }
	
	// Then 
	theGame.processKeyDown();

	// Depending on the game may need a keylistener for keyup and keypress also
};
// ----------------------------------------------------------------------------
function doTouchStart(e)
{
    theGame.processKeyDown();
    
    e.preventDefault(); // disable browser's default behaviour and prevent multiple event triggers
    return false;		
};


// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
function doKeyUp(e)
{
    //console.log("DEBUG: doKeyUP entered")
	theGame.processKeyUp();
};
// ----------------------------------------------------------------------------
function doTouchEnd(e)
{
    theGame.processKeyUp();
    
    e.preventDefault(); // disable browser's default behaviour and prevent multiple event triggers
    return false;		
};


// ----------------------------------------------------------------------------
//                                                        window.ONLOAD
// ----------------------------------------------------------------------------
window.onload = function()
{
	// Size all canvas to correct Ratio before doing anything 
	// Needed for resizing and centring to work correctly from the start
	doResize();
	
	// Initialize the game
	theGame.Init();

	// Run the game (start the game loop)
	runGame();
};

// ----------------------------------------------------------------------------
//                                                         the game 'loop'
// Event handlers/listeners cover the game input
// ----------------------------------------------------------------------------
function runGame()
{
	//theGame.Draw();    TODO - implement          
	theGame.Update();

    // Be certain to call runGame again	
	requestAnimationFrame(runGame);
};

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// end of file
